------------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand  
-- 
-- Create Date:    18:07:09 11/11/2015 
-- Design Name: 
-- Module Name:    Datapath - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Datapath is
  PORT(
    clk_i : in std_logic;
    --Input
    data_i    : in std_logic_vector(7 downto 0);
    key_i     : in std_logic_vector(7 downto 0);
    --Control
    mode_i    : in std_logic;
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(2 downto 0);
    Z_sel     : in std_logic_vector(1 downto 0);
    L2_sel    : in std_logic_vector(1 downto 0);
    count_r   : in std_logic_vector(4 downto 0);
    count_i   : in std_logic_vector(2 downto 0);
    --Output
    data_o : out std_logic_vector(7 downto 0)
  );
end Datapath;

architecture Behavioral of Datapath is

  COMPONENT keyschedule
	PORT(
		key_in	: in	std_logic_vector(79 downto 0);
		mode_i	: in std_logic;
		key_o_e	: out	std_logic_vector(79 downto 0);
		key_o_d	: out	std_logic_vector(79 downto 0)
		);
	END COMPONENT;

  COMPONENT extKey
	PORT(
    Z       : IN std_logic_vector(3 downto 0);
		i       : IN std_logic_vector(4 downto 0);
		counter : IN std_logic_vector(2 downto 0);          
		rk_o    : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  COMPONENT permutation_layer
	PORT(
		mode_i  : IN std_logic;
		state_i : IN std_logic_vector(63 downto 0);          
		state_o : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT round_f
	PORT(
    nible_l1_i : IN std_logic_vector(3 downto 0);
		nible_r_i  : IN std_logic_vector(3 downto 0);
		rk         : IN std_logic_vector(3 downto 0);
		nible_o    : OUT std_logic_vector(3 downto 0)
		);
	END COMPONENT;
  
  component xor4to4 is
    Port ( nible_1_i : in  STD_LOGIC_VECTOR (3 downto 0);
           nible_2_i : in  STD_LOGIC_VECTOR (3 downto 0);
           nible_o : out  STD_LOGIC_VECTOR (3 downto 0));
  end component;
  
  --signals
  --regsiters
  signal key_reg    : std_logic_vector(key_size-1 downto 0);
  signal state_reg  : std_logic_vector(state_size-1 downto 0);
  
  signal L_o            : std_logic_vector(3 downto 0);
  signal rk             : std_logic_vector(3 downto 0);
  signal round_o        : std_logic_vector(3 downto 0);
  signal L_i            : std_logic_vector(3 downto 0);
  signal L2_i           : std_logic_vector(3 downto 0);
  signal acc            : std_logic_vector(3 downto 0);
  signal Z              : std_logic_vector(31 downto 0);
  signal perm_i         : std_logic_vector(state_size-1 downto 0);
  signal perm_o         : std_logic_vector(state_size-1 downto 0);
  signal k_s_o_1, k_s_o_2          : std_logic_vector(key_size-1 downto 0);
  
begin
  --key process
  K_proc : process(clk_i, key_sel)
  begin
    if rising_edge(clk_i) then
      if key_sel = "01" then
        key_reg <= key_i & key_reg(key_size-1 downto 8);
      elsif key_sel = "10" then
        key_reg <= k_s_o_1;
		elsif key_sel = "11" then
        key_reg <= k_s_o_2;
      else
        key_reg <= key_reg;
      end if;
    end if;
  end process;
  
  --data process
  D_proc : process(clk_i, data_sel)
  begin
    if rising_edge(clk_i) then
      if data_sel = "001" then
        state_reg <= data_i & state_reg(state_size-1 downto 8);
      elsif data_sel = "010" then
        state_reg(63 downto 32) <= L_o & state_reg(63 downto 36);
        state_reg(31 downto 0)  <= state_reg(27 downto 0) & state_reg(31 downto 28);
      elsif data_sel = "011" then
        state_reg <= perm_o;
      elsif data_sel = "100" then
        state_reg <= state_reg;
      else
        state_reg <= (others => '0');
      end if;
    end if;
  end process;
  
  data_o <= state_reg(7 downto 0); --when done_s = '1' else x"00";

  -- cipher component
  --keyschedule
  Inst_keyschedule: keyschedule PORT MAP(
		key_in => key_reg,
		mode_i => mode_i,
		key_o_e  => k_s_o_1,
		key_o_d  => k_s_o_2
	);

  Inst_extKey: extKey PORT MAP(
		Z       => Z(3 downto 0),
		i       => count_r,
    counter => count_i,
		rk_o    => rk
	);

  Z_proc : process(clk_i, Z_sel)
  begin
    if rising_edge(clk_i) then
      if Z_sel = "00" then
        Z(3 downto 0) 	<= key_reg(64) & key_reg(40) & key_reg(24) & key_reg(4);
        Z(7 downto 4) 	<= key_reg(65) & key_reg(41) & key_reg(25) & key_reg(5);
        Z(11 downto 8) 	<= key_reg(66) & key_reg(42) & key_reg(26) & key_reg(6);
        Z(15 downto 12) <= key_reg(67) & key_reg(43) & key_reg(27) & key_reg(7);
        Z(19 downto 16) <= key_reg(72) & key_reg(52) & key_reg(36) & key_reg(12);
        Z(23 downto 20) <= key_reg(73) & key_reg(53) & key_reg(37) & key_reg(13);
        Z(27 downto 24) <= key_reg(74) & key_reg(54) & key_reg(38) & key_reg(14);
        Z(31 downto 28) <= key_reg(75) & key_reg(55) & key_reg(39) & key_reg(15);
      elsif Z_sel = "11" then
        Z <= Z(31 downto 28) & Z(31 downto 4);
      else
		  if mode_i = '0' then
			  Z(3 downto 0)  <= k_s_o_1(64) & k_s_o_1(40) & k_s_o_1(24) & k_s_o_1(4);
			  Z(7 downto 4)  <= k_s_o_1(65) & k_s_o_1(41) & k_s_o_1(25) & k_s_o_1(5);
			  Z(11 downto 8) <= k_s_o_1(66) & k_s_o_1(42) & k_s_o_1(26) & k_s_o_1(6);
			  Z(15 downto 12)<= k_s_o_1(67) & k_s_o_1(43) & k_s_o_1(27) & k_s_o_1(7);
			  Z(19 downto 16)<= k_s_o_1(72) & k_s_o_1(52) & k_s_o_1(36) & k_s_o_1(12);
			  Z(23 downto 20)<= k_s_o_1(73) & k_s_o_1(53) & k_s_o_1(37) & k_s_o_1(13);
			  Z(27 downto 24)<= k_s_o_1(74) & k_s_o_1(54) & k_s_o_1(38) & k_s_o_1(14);
			  Z(31 downto 28)<= k_s_o_1(75) & k_s_o_1(55) & k_s_o_1(39) & k_s_o_1(15);
			else
			  Z(3 downto 0)  <= k_s_o_2(64) & k_s_o_2(40) & k_s_o_2(24) & k_s_o_2(4);
			  Z(7 downto 4)  <= k_s_o_2(65) & k_s_o_2(41) & k_s_o_2(25) & k_s_o_2(5);
			  Z(11 downto 8) <= k_s_o_2(66) & k_s_o_2(42) & k_s_o_2(26) & k_s_o_2(6);
			  Z(15 downto 12)<= k_s_o_2(67) & k_s_o_2(43) & k_s_o_2(27) & k_s_o_2(7);
			  Z(19 downto 16)<= k_s_o_2(72) & k_s_o_2(52) & k_s_o_2(36) & k_s_o_2(12);
			  Z(23 downto 20)<= k_s_o_2(73) & k_s_o_2(53) & k_s_o_2(37) & k_s_o_2(13);
			  Z(27 downto 24)<= k_s_o_2(74) & k_s_o_2(54) & k_s_o_2(38) & k_s_o_2(14);
			  Z(31 downto 28)<= k_s_o_2(75) & k_s_o_2(55) & k_s_o_2(39) & k_s_o_2(15);
			end if;
      end if;
    end if;
  end process;
  
  --non linear layer
  Inst_round_f: round_f PORT MAP(
    nible_l1_i => state_reg(35 downto 32),
		nible_r_i  => state_reg(31 downto 28),
		rk         => rk,
		nible_o    => round_o
	);
  
  --linear layer
  L : xor4to4 port map(
  nible_1_i => round_o,
  nible_2_i => L_i,
  nible_o   => L_o
  );
  
  L_i <= acc when count_i="111" else L2_i;
  L_i_process : process(clk_i, L2_sel)
  begin
  if rising_edge(clk_i) then
    if L2_sel="01" then
      L2_i  <= state_reg(31 downto 28);
      acc   <= state_reg(31 downto 28);
    elsif L2_sel="11" then
      L2_i  <= L2_i;
      acc   <= acc xor state_reg(31 downto 28);
    else
      L2_i  <= x"0";
      acc   <= x"0";
    end if;
  end if;
  end process;
  
  --permutation layer
  Inst_permutation_layer: permutation_layer PORT MAP(
		mode_i  => mode_i,
		state_i => perm_i,
		state_o => perm_o
	);
	
  perm_i <= L_o & state_reg(63 downto 36) & state_reg(27 downto 0) & state_reg(31 downto 28);
  
end Behavioral;

